;-------------------------------------------------------------------------------
; * Generic File Function Library
; Version 05 Aug 2012
; Created by Teasy <TheChange@yahoo.com>
;-------------------------------------------------------------------------------
; This library contains the following functions:
;
;   FullPath$( path$ )                  Convert relative to absolute path
;   ValidateFilename( Filename$ )       Make sure just file name is valid
;   ValidateFoldername( Foldername$ )   Make sure entire path name is valid
;   ConstructFolder( Foldername$ )      Create cascaded directory chain
;   WriteText( FileHandle , Message$ )  Write ASCII text to open file
;   ReadText$( FileHandle , Length )    Read ASCII text from open file
;   MoveFile$( Source$ , Target$ )      Rename or move a file or folder
;   FindFreeFileName$( Folder$ , Pattern$ )  Handy for unique temp/mass files
;
; Also see the SHFileOp.Lib.BB file
; to implement the Windows shell copy process and dialog
; greatly simplifying many copy/move/delete/rename operations.
;-------------------------------------------------------------------------------
; 05 Aug 2012
; - Added some more comments (revised comments related to string lib)
; 07 Aug 2010
; - Modified MoveFile error response to be more uniform.
; - Modified MoveFile comments for improved clarity.
; 01 Apr 2007
; - Last known modification.
;-------------------------------------------------------------------------------
; Uses String Library
;
;   Include "\Dev\Blitz\Lib\String.Lib.BB"
;
; Which also contains the following file/folder functions:
;
;   String_TrailingBackslash$( path$ )
;       Add trailing backslash, if not present
;
;   String_RemoveTrailingBackslash$( path$ )
;       Wipe trailing backslash, if present
;
;   String_GetFileName$( LookIn$ , Separator$="\" )
;       Keep only file name, strip path and drive (if used)
;       - Cuts everything before last "\" or first ":" respectively
;
;   String_GetPathName$( FileSpec$ , Separator$="\" )
;       Strip file name from absolute file location
;       - Cuts everything after last "\" or first ":" respectively
;
;   String_RemoveExtension$( name$ )
;       Removes (trailing) extension from filename (if any)
;       - Can be called repeatedly to remove multiple extensions
;       - May include path (and will not be removed)
;
;   String_GetFileExtension$( name$ )
;       Removes (trailing) extension from filename (if any)
;       - May include path (and will not be removed)
;
;-------------------------------------------------------------------------------


; Return full (absolute) path from a relative path
Function FullPath$ ( path$ )
    Local Current$
    Local FirstChar$
    Local Result$

    ; check if path already absolute (contains drive)
    If Instr( path , ":" )
        Result = path
    ; without drive
    Else
        ; remember current drive and path
        Current = CurrentDir()
        ; check if path is a root location (absolute without drive)
        FirstChar = Left( path , 1 )
        If FirstChar = "\" Or FirstChar = "/"
            ; current drive + root location
            Result = Left( Current , 2 ) + path
        ; not a root location (relative)
        Else
            ; current drive, folder + relative path
            Result = String_TrailingBackslash( Current ) + path
        EndIf
    EndIf
    ; use a trailing backslash if resulting (absolute) path is a folder
    If FileType( Result ) = 2
        Result = String_TrailingBackslash( Result )
    EndIf

    Return Result
End Function


; Fierce file name validation
Function ValidateFilename ( Filename$ )
    Local Charnum
    Local Length
    Local Char$
    Local InvCharnum

    ; no be leading or trailing spaces (uses custom Trim to support extended chars)
    Filename = String_Trim( Filename )
    ; hunt for invalid chars: "*/:<>?\|" and control characters (0..31)
    Length = Len( Filename )
    For Charnum = 1 To Length
        Char = Mid( Filename , Charnum,1 )
        ; control character
        If Asc( Char ) < 32
            ; abort instantly
            Return False
        ; normal or extended character
        Else
            ; invalid (Windows) filename characters
            If Char="*" Or Char="/" Or Char=":" Or Char="<" Or Char=">" Or Char="?" Or Char="\" Or Char="|"
                Return False
            EndIf
        EndIf
    Next

    ; passed!
    Return True
End Function


; Fierce folder name validation
Function ValidateFoldername ( Foldername$ )
    Local Letter
    Local Segment,Segments
    Local Check$

    ; replace slashes with backslashes
    Foldername = Replace( Foldername , "/" , "\" )

    ; * validate drive (if present)

    ; optional leading drive name format: "x:\"
    Select String_CountChars( Foldername , ":" )
        Case 1
            ; if only 1 colon, it must be on position 2
            If Not Mid( Foldername , 2,1 ) = ":"
                Return False
            EndIf
            ; and drive letter must be valid on position 1
            Letter = Asc( Upper( Left( Foldername , 1 ) ) )
            If Letter < 65 Or Letter > 90  ; below A or above Z
                Return False
            EndIf
            ; clip drive (no longer required for validation)
            Foldername = Right( Foldername , Len( Foldername ) - 2 )
        Case 2
            ; multiple colons?!
            Return False
    End Select

    ; * now iterate through backslashes
    ;   and run each name through the validate filename routine

    ; check number of sub-folders included
    Segments = String_ParameterCount( Foldername , "\" )
    For Segment = 1 To Segments
        ; next backslash separated portion
        Check = String_Parameter( Foldername , Segment , "\" )
        ; validate it
        If Not ValidateFilename( Check )
            ; invalid
            Return False
        EndIf
    Next

    ; passed!
    Return True
End Function


; Constructs a single chain of nested folders, e.g. "a\b\c"
Function ConstructFolder ( Foldername$ )
    Local Segment,Segments
    Local Separator
    Local FileType
    Local Construct$

    ; using only backslashes for easy parameter tracking
    Foldername = Replace( Foldername , "/" , "\" )
    ; number of sub-folders included
    Segments = String_ParameterCount( Foldername , "\" )
    ; clip drive (if present)
    If Instr( Foldername , ":" )
        Foldername = Right( Foldername , Len( Foldername ) - 2 )
    EndIf
    ; start iterating at 1st character (1st backslash will be skipped)
    Separator = 1
    For Segment = 1 To Segments
        ; find next backslash
        Separator = Instr( Foldername , "\" , Separator+1 )
        ; in case this was the last (or none at all)
        If Separator <= 0
            ; take full (relative) path as construction path
            Construct = Foldername
        ; not yet the last backslash
        Else
            ; select all of (relative) path up to next backslash
            Construct = Left( Foldername , Separator-1 )
        EndIf
        ; make sure the segment to construct isn't a file
        FileType = FileType( Construct )
        If FileType = 1
            ; there's nothing construct from this point,
            ; (unless you remove the file :P)
            Return False
        ; if the folder already exists
        ElseIf FileType = 2
            ; no need to create it
        ; no file or folder with that name, yet
        ElseIf FileType = 0
            ; create it
            CreateDir Construct
        EndIf
    Next

    Return True
End Function


; Write a string (without header) to current position of file
Function WriteText( FileHandle , Message$ )
    Local Char
    Local Length
    Local Buffer

    Length = Len( Message )
    Buffer = CreateBank( Length )
    ; convert string to bank
    For Char = 1 To Length
        PokeByte Buffer , Char-1 , Asc( Mid( Message , Char,1 ) )
    Next
    ; then write bank to file (single write is faster)
    WriteBytes Buffer,FileHandle , 0,Length
    FreeBank Buffer
End Function


; Read a string (without header) from current position of file
Function ReadText$( FileHandle , Length )
    Local Char
    Local Buffer
    Local Message$

    Buffer = CreateBank( Length )
    ; read block of data from file and put into bank
    ReadBytes Buffer,FileHandle , 0,Length
    ; then convert bank to string
    For Char = 1 To Length
        Message = Message + Chr( PeekByte( Buffer , Char-1 ) )
    Next
    FreeBank Buffer

    Return Message
End Function


;-------------------------------------------------------------------------------
Function MoveFile$( Source$ , Target$ )
;---------------------------------------------------------------------------
; Version 07 Aug 2010 Revision 1
;---------------------------------------------------------------------------
; Rename or move a file or folder
;---------------------------------------------------------------------------
; Uses API_MoveFile:
;   file:   kernel32.decls
;   contents:   .lib "KERNEL32.DLL"
;               API_MoveFile%(source$,target$) : "MoveFileA"
;---------------------------------------------------------------------------
; MoveFile always returns extended error info, with this format:
;       Error code (2), Space, Colon (:), Space, Error message
; You can grab the unique error code using:
; - Automatic typecast:
;       ErrorCode% = MoveFile( Source , Target )
; - Manual typecast:
;       If Int MoveFile( Source , Target ) = True Then Print "Woo!"
; You can grab the error message using:
;       Result$ = MoveFile( Source , Target )
;       ErrorMsg$ = Mid( Result , 6 )
;       ;ErrorCode% = Int Result
;---------------------------------------------------------------------------
; List of extended error info:
;        1 : Operation succesful
;        0 : Unhandled exception
;       -1 : Target already exists
;       -2 : Source does not exist
;       -3 : Unable to move Source to Target
;       -4 : Target not created
;       -5 : Source not moved
;       -6 : Source not specified
;       -7 : Target not specified
;---------------------------------------------------------------------------
    Source = Trim( Source )
    Target = Trim( Target )
    If Source = ""
        Return "-6 : Source not specified"
    EndIf
    If Target = ""
        Return "-7 : Target not specified"
    EndIf
    ; Target already exists
    If FileType( Target ) <> 0
        Return "-1 : Target already exists"
    EndIf
    ; Source does not exist
    If FileType( Source ) = 0
        Return "-2 : Source does not exist"
    EndIf
    ; Attempt to move source to target
    If api_MoveFile( Source , Target ) = 0
        Return "-3 : Unable to move source to target"
    EndIf
    ; Source removed, target created
    If FileType( Source ) = 0 And FileType( Target ) <> 0
        Return " 1 : Operation succesful"
    Else
        ; Target not created
        If FileType( Target ) = 0
            Return "-4 : Target not created"
        EndIf
        ; Source not moved
        If FileType( Source ) <> 0
            Return "-5 : Source not moved"
        EndIf
        Return " 0 : Unhandled exception"
    EndIf
;-------------------------------------------------------------------------------
End Function
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
; Generate a non-existing filename in specific folder
; - using the current timestamp
; - with optional filename pattern
;-------------------------------------------------------------------------------
Function FindFreeFileName$( Folder$ , Pattern$="*" )
;-------------------------------------------------------------------------------
;   Param Folder$   ; where to look for the unexisting file
;   Param Pattern$  ; what the file looks like, format:
;                   ; * = timestamp here (YYYYMMDD.HHMMSS[.X])
;                   ; example: "Temp.*.txt"
;                   ; notes: only when the file already exists without extended number (.X)
;                   ;        this extra number is added, starting at 1, counting upwards.
;                   ;        when the extra number reaches 999, random (alpha)chars will be used.
;-------------------------------------------------------------------------------
    Local filename$
    Folder = String_TrailingBackslash(Folder)
    Local attempts = 0
    
    ; attempt 1: simple timestamp
    Local timestamp$ = String_CurrentTimestamp(0)  ; YYYYMMDD.HHMMSS
    filename = Folder + String_Replace(Pattern,"*",timestamp)
    If FileType(filename) = 0 Then Return filename
    attempts=attempts+1
    
    ; attempt 2: added number (extended loop)
    ; - can be improved using '1048576 algorithm'
    ;   e.g. 1y -> 2y -> 4y -> 8y -> 16y -> 32n -> 24y -> 28n -> 26y -> 27n
    ;   although this can leave gaps in the numbered files
    ;   if not all numbered files are used in the first place.
    Local maxnumber=999
    Local number
    For number = 1 To maxnumber
        filename = Folder + String_Replace(Pattern,"*",timestamp+"."+number)
        If FileType(filename) = 0 Then Return filename
        attempts=attempts+1
    Next
    
    ; attempt 3: added random chars (extended loop 2)
    ; - random chars are alpha only (A..Z)
    Local stagechange=10  ; x attempts per stage: before adding another char
    Local count=1
    Local length=1
    Repeat
        filename = Folder + String_Replace(Pattern,"*",timestamp+"."+String_GenerateGarbage(length,65,90))
        If FileType(filename) = 0 Then Return filename
        attempts=attempts+1
        count=count+1
        If count>=stagechange
            count=1
            length=length+1
            ; stop at x chars
            If length>=99
                DebugLog "FindFreeFileName( "+Folder+" , "+Pattern+" ): Unable to find a free filename (after "+attempts+" attempts)"
                Return
            End If
        End If
    Forever
;-------------------------------------------------------------------------------
End Function
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
; Return the full filepath of the (active) host executable (process)
;-------------------------------------------------------------------------------
Function ExecutableFilePath$()
;-------------------------------------------------------------------------------
; .lib "KERNEL32.DLL"
; api_GetModuleFileName% (module_handle, fname*, fname_size) : "GetModuleFileNameA"
;-------------------------------------------------------------------------------
    Local PathSize
    Local Bank
    Local Length
    Local FilePath$
    Local Byte
    
    PathSize = 260
    Bank = CreateBank( PathSize )
    ; module 0 means calling module (e.g. host)
    Length = api_GetModuleFileName( 0 , Bank , PathSize )
    
    If Length
        Length=Length-1
        For Byte = 0 To Length ;Length-1
            FilePath = FilePath + Chr( PeekByte( Bank , Byte ) )
        Next
    EndIf
    
    FreeBank Bank
    
    Return FilePath
;-------------------------------------------------------------------------------
End Function
;-------------------------------------------------------------------------------

;~IDEal Editor Parameters:
;~C#Blitz3D1105