;-------------------------------------------------------------------------------
; * Mathematical Routines
; Ver 2010 Feb 12
;-------------------------------------------------------------------------------
; This library auto-initializes
; Note: Some of these functions are NOT backward compatible with pre-2007 code
;-------------------------------------------------------------------------------
; * Change-Log
; 21 Aug 2012
; - Added IntervalInt, a simplified variation of InterleaveInt
; - And also variants Low/High if evenly spaced values are needed (Floor/Ceil)
; 2010 Feb 12
; - Localized variables in InterceptAngle function
; 2010 Feb 26
; - optimized ScalePoint (30% faster)
; - added ScalePointTri
; - added random generator encapsulation routines
; 2011 Nov 03
; - added PointOnRect to include the border too (PointInRect ignores border)
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
; Random Encapsulation (pre-cached blitz random generator)
;-------------------------------------------------------------------------------

Const randencap_defaultdiversity = 73451
Global randencap_position = 0
Global randencap_diversity = 0
Dim randencap_value#( 0 )

Function Random#( startpoint# = 0 , endpoint# = 1 )
	randencap_position = randencap_position + 1
	If randencap_position > randencap_diversity
		randencap_position = 0
	EndIf
	Return startpoint + (endpoint - startpoint) * randencap_value( randencap_position )
End Function

Function InitRandom( seed = -2147483648 , diversity = randencap_defaultdiversity )
    ; seed generator
	If seed = -2147483648
		seed = MilliSecs()
	EndIf
	SeedRnd seed
    ; fill array
    randencap_diversity = diversity
    Dim randencap_value( diversity )
	For randencap_position = 0 To randencap_diversity
		randencap_value( randencap_position ) = Rnd( 0,1 )
	Next
    ; from the top
	randencap_position = 0
End Function


;-------------------------------------------------------------------------------
; Precalculate look-up tables
;-------------------------------------------------------------------------------

Const Math_DistMax = 100
Dim Math_DistTable( Math_DistMax,Math_DistMax )

Global Math_SinTable#[ 360 ]
Global Math_CosTable#[ 360 ]

Function Math_InitTables()
    Local x,y
    Local Angle
    
    For y = 0 To Math_DistMax
        For x = 0 To Math_DistMax
            Math_DistTable( x,y ) = Sqr( x*x + y*y )
        Next
    Next
    
    For Angle = 0 To 360
        Math_SinTable[ Angle ] = Sin( Angle )
        Math_CosTable[ Angle ] = Cos( Angle )
    Next 
End Function

Math_InitTables


;-------------------------------------------------------------------------------
; Wrapper functions for look-up tables
; - very fast for imprecise trigonometry
;-------------------------------------------------------------------------------

; - accuracy <~1% on large numbers
; - slower than VectorDist! (probably due to 2 divisions)
Function qDist( dx,dy )
    Local scale
    
    ; make positive
    If dx < 0
        dx = -dx
    EndIf
    If dy < 0
        dy = -dy
    EndIf
    ; out of table range
    If dx > Math_DistMax Or dy > Math_DistMax
        ; most significant dimension
        If dx > dy
            ; scale to fit in table (down & up again)
            scale = dx / Math_DistMax
            Return Math_DistTable( Math_DistMax,dy/scale ) * scale
        Else
            ; scale to fit in table (down & up again)
            scale = dy / Math_DistMax
            Return Math_DistTable( dx/scale,Math_DistMax ) * scale
        EndIf
    Else
        ; straight pass through
        Return Math_DistTable( dx,dy )
    EndIf
    
End Function


Function qSin#( Value% )
    
    If Value <    0 Then Return Math_SinTable[ 360 - Abs( Value ) Mod 360 ]
    If Value >= 360 Then Return Math_SinTable[ Value Mod 360 ]
    
    Return Math_SinTable[ Value ]
End Function 


Function qCos#( Value% )
    
    If Value <    0 Then Return Math_CosTable[ 360 - Abs( Value ) Mod 360 ]
    If Value >= 360 Then Return Math_CosTable[ Value Mod 360 ]
    
    Return Math_CosTable[ Value ]
End Function 


;-------------------------------------------------------------------------------
; Vectors / trigonometry
;-------------------------------------------------------------------------------

; polar & cartesian coordinate conversion routines
Function VectorX# ( Distance# , Angle# )    Return Sin ( Angle      ) * Distance    End Function
Function VectorY# ( Distance# , Angle# )    Return Sin ( Angle - 90 ) * Distance    End Function
Function VectorDistance# ( X# , Y# )        Return Sqr ( X * X + Y * Y )            End Function
Function VectorAngle#    ( X# , Y# )        Return -ATan2 ( X , Y ) + 180           End Function

; short hand
Function vx# ( d# , a# )    Return Sin ( a ) * d            End Function
Function vy# ( d# , a# )    Return Sin ( a - 90 ) * d       End Function
Function vd# ( x# , y# )    Return Sqr ( x * x + y * y )    End Function
Function va# ( x# , y# )    Return -ATan2 ( x , y ) + 180   End Function

; quick and imprecise
Function qvx# ( d# , a# )    Return qSin ( a ) * d          End Function
Function qvy# ( d# , a# )    Return qSin ( a - 90 ) * d     End Function


;-------------------------------------------------------------------------------
; Range limiting (overflow)
;-------------------------------------------------------------------------------

Function RangeFloat#( Value# , Max#=360 )
    
    If Value <    0 Then Return Max - Abs( Value ) Mod Max
    If Value >= Max Then Return Value Mod Max
    
    Return Value
End Function 


Function RangeInt%( Value% , Max%=360 )
    
    If Value <    0 Then Return Max - Abs( Value ) Mod Max
    If Value >= Max Then Return Value Mod Max
    
    Return Value
End Function 


;-------------------------------------------------------------------------------
; Capping (no overflow)
;-------------------------------------------------------------------------------

Function CapFloat#( Value# , Min#,Max# )
    
    If Value < Min Then Return Min
    If Value > Max Then Return Max
    
    Return Value
End Function


Function CapInt%( Value% , Min%,Max% )
    
    If Value < Min Then Return Min
    If Value > Max Then Return Max
    
    Return Value
End Function


;-------------------------------------------------------------------------------
; Comparison
;-------------------------------------------------------------------------------

Function HighestFloat#( Value1# , Value2# )
    
    If Value1 > Value2
        Return Value1
    Else 
        Return Value2
    EndIf 
    
End Function


Function HighestInt%( Value1% , Value2% )
    
    If Value1 > Value2
        Return Value1
    Else 
        Return Value2
    EndIf 
    
End Function


Function LowestFloat#( Value1# , Value2# )
    
    If Value1 < Value2
        Return Value1
    Else 
        Return Value2
    EndIf 
    
End Function


Function LowestInt%( Value1% , Value2% )
    
    If Value1 < Value2
        Return Value1
    Else 
        Return Value2
    EndIf 
    
End Function


;-------------------------------------------------------------------------------
; Step limiting (interleaving)
;-------------------------------------------------------------------------------

Function InterleaveInt%( Angle% , RotationInterval%=5,Angles%=360 )
    ; note: rounds to nearest angle, causing some fluxuations (e.g. some angles occur more often)
    ; for evenly used angles, use: RangeInt( IntervalInt(Angle,Angles) , RotationInterval )
    Return Int( Float( RangeInt( Angle,Angles ) - RotationInterval ) / RotationInterval ) * RotationInterval + RotationInterval
End Function


Function IntervalInt( Value% , Interval% )
    ; note: rounds to nearest value, causing some fluxuations (e.g. some values occur more often)
    ; for evenly used value, used the Low or High variant of this function
    Return Int( Float( Value - Interval ) / Interval ) * Interval + Interval
End Function

Function IntervalIntLow( Value% , Interval% )
    ; note: rounds to lowest value, causing no fluxuations (all values occur evenly)
    Return Floor( Float( Value - Interval ) / Interval ) * Interval + Interval
End Function

Function IntervalIntHigh( Value% , Interval% )
    ; note: rounds to highest value, causing no fluxuations (all values occur evenly)
    Return Ceil( Float( Value - Interval ) / Interval ) * Interval + Interval
End Function


;-------------------------------------------------------------------------------
; Scaling
;-------------------------------------------------------------------------------

Function InterpolatePoint#( start#,finish# , delta#=0.5 )
    Return start + (finish-start) * delta
End Function


; scale a point from range1 to range2
Function ScalePoint#( point# , range1start#,range1end# , range2start#,range2end# )
    Return (point-range1start) / (range1end-range1start) * (range2end-range2start) + range2start
End Function


; like scalepoint, but with a controllable middle node (dual-linear interpolation)
Function ScalePointTri#( p# , s1#,e1# , s2#,e2# , k#=0.0 )
    ; point, start1,end1, start2,end2
    ; kneepoint (-1 = lowest, 0 = center, 1 = highest)
    
    ; not fully optimized
    ; but speed is okay
    
    If k = 0
        ;Return ScalePoint( p , s1,e1 , s2,e2 )
        Return (p-s1) * (e2-s2) / (e1-s1) + s2
    End If
    
    ; range1, range2, delta
    Local r1# = e1-s1
    Local r2# = e2-s2
    Local d# = (p-s1) / r1
    
    ; segment 1 (left)
    If d < 0.5
        ; 0 .. 0          : k = -1
        ; 0 .. r2/2 [+s2] : k =  0
        ; 0 .. r2         : k =  1
        ;r2 = r2 * 2
        d = d * 2
        k = (k + 1) / 2
        Return r2 * d * k + s2
    ; segment 2 (right)
    Else
        ; 0    .. r2       : k = -1
        ; r2/2 .. r2 [+s2] : k =  0
        ; r2   .. r2       : k =  1
        ;d = d*2
        ;k = (k + 1) / 2
        ;r2 = r2 * 2
        ;Return ScalePoint( d * k , 0.5+0,1+1 , 0,r2 )
        d = 1 - (d - 0.5) * 2  ; 0..1
        k = 1 - (k + 1) / 2  ; 0..1
        ;Return r2-ScalePoint( (1-d)*(1-k) , 0,1 , 0,r2 )+s2
        ;Return ScalePoint( d*k , 0,1 , r2,0 )+s2
        Return r2 - (d*k) * r2 + s2
    End If
End Function


;-------------------------------------------------------------------------------
; Angles
;-------------------------------------------------------------------------------

; rotate 2D coordinate around point (0,0)
Global RotatedX#
Global RotatedY#
Function RotatePoint( x#,y# , a# )
    Local d# = Sqr( x*x + y*y )
    a = a + 180 - ATan2( x,y )
    RotatedX = Sin( a    ) * d
    RotatedY = Sin( a-90 ) * d
End Function


; return relative angle difference (takes angle wrapping into account)
; - max angle difference: 359 degrees
Function AngleDifference( CurrentAngle , TargetAngle )
    
    If TargetAngle - CurrentAngle > 180 Then TargetAngle = TargetAngle - 360
    If CurrentAngle - TargetAngle > 180 Then CurrentAngle = CurrentAngle - 360
    
    Return CurrentAngle - TargetAngle
End Function


; calculate intelligent hit angle with constant object and projectile speed
Const Math_InterceptAngle_OutOfRange = -1001
Function InterceptAngle( ex#,ey# , evx#,evy# , px#,py# , pvx#,pvy# , bs# , maxtime=0 )
    ;---------------------------------------------------------------------------;-----------------------------------
    ;   Find projectile intercept angle from Source (Enemy) to Target (Player)  ; Algorithm: Eelco <eelco_hoogendoorn@hotmail.com>
    ;---------------------------------------------------------------------------; Wrapper: Teasy <thechange@yahoo.com>
    ;   Parameters                                                              ; Source: INTER~29.BB
    ;---------------------------------------------------------------------------------------------------------------
    ;   (ex,ey)     Enemy position
    ;   evx & evy   Enemy velocity vector
    ;   (px,py)     Player position
    ;   pvx & pvy   Player velocity vector
    ;   bs          Bullet speed
    ;   maxradius   Radius for the player to be in (now or future), 0 = no limit
    ;   maxtime     Only shoot if (future) time of impact is below this amount
    ;               (0 = no limit)
    ;---------------------------------------------------------------------------
;
    ;---------------------------------------------------------------------------
    ;   Make player motion relative to enemy motion
    ;---------------------------------------------------------------------------
;
    Local es# = Sqr( evx*evx + evy*evy )
    Local ea# = ATan2( evy,evx )
;
    Local ps# = Sqr( pvx*pvx + pvy*pvy )
    Local pa# = ATan2( pvy,pvx )
;
    Local rpvx#=pvx-evx                                                         ; Relative player velocity
    Local rpvy#=pvy-evy                                                         ; (Cart vector)
;
    ps=Sqr( rpvx*rpvx + rpvy*rpvy )                                             ; Relative player motion (polar) vector
    pa=ATan2( -rpvy,rpvx )                                                      ; Warning: overrides ps,pa
;
    ;---------------------------------------------------------------------------
    ;   Relative distance to player from enemy
    ;---------------------------------------------------------------------------
;
    Local rx#=px-ex                                                             ; Relative player-enemy distance
    Local ry#=py-ey                                                             ; (Cart vector)
;
    ;---------------------------------------------------------------------------
    ;   Player must be moving
    ;---------------------------------------------------------------------------
;
    If Abs( ps )>=0.000001                                                      ; If player speed is more than a little
;   :
        ;-----------------------------------------------------------------------
        ;   Find where player will be, using player motion and bullet speed
        ;-----------------------------------------------------------------------
;   :
        Local a#=ATan2( ey-py,ex-px )                                           ; Angle player->enemy
        Local n#=Sin( -pa-a )/bs*ps                                             ; Angle relative to player, speed scaled
;   :
        ;-----------------------------------------------------------------------
        ;   Player path must be within range for bullet to hit
        ;-----------------------------------------------------------------------
;   :
        If n>=-1 And n<=1                                                       ; Can bullet ever hit player?
;   :   ;
            ;-------------------------------------------------------------------
            ;   New bullet motion for impact
            ;-------------------------------------------------------------------
;   :   ;
            Local ba#=180+ASin( n )-a                                           ; Rotated angle back to enemy
;   :   ;
            Local bvx#=Cos( ba )*bs                                             ; New bullet velocity (cart) vector
            Local bvy#=-Sin( ba )*bs
;   :   ;
            ;-------------------------------------------------------------------
            ;   See when the bullet will impact
            ;-------------------------------------------------------------------
;   :   ;
            Local d#=bvy*pvx - bvx*pvy                                          ; Intersection
;   :   ;
            If Abs( d )>=0.000001                                               ; Avoid infinitely small
                Local t#=( bvx*ry - bvy*rx )/d                                  ; Time of intersection
;   :   ;   |
                If Not ( maxtime>0 And t>=maxtime Or t<0 )                      ; Time must be in future but not too far
                    Return VectorAngle( bvx,bvy )                               ; Convert back to: Up = 0 degrees
                EndIf
;   :   ;   |
            EndIf
;   :   ;
        EndIf
    EndIf
;
    Return Math_InterceptAngle_OutOfRange
End Function


;-------------------------------------------------------------------------------
; Collision
;-------------------------------------------------------------------------------

; really INside rect, e.g. do not count the 'border' of the rect
Function PointInRect( px,py , rx,ry , rw,rh )
    
    If px>rx
        If py>ry
            If px<rx+rw-1
                If py<ry+rh-1
                    Return True
                EndIf
            EndIf
        EndIf
    EndIf
    
End Function

; same as PointInRect, but DO count the border
Function PointOnRect( px,py , rx,ry , rw,rh )
    Return px>=rx And py>=ry And px<=rx+rw-1 And py<=ry+rh-1
End Function

Function PointInTriangle( x0,y0 , x1,y1 , x2,y2 , x3,y3 )
    Local b0#,b1#,b2#,b3#
    
    b0 =  (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1)
    b1 = ((x2 - x0) * (y3 - y0) - (x3 - x0) * (y2 - y0)) / b0 
    b2 = ((x3 - x0) * (y1 - y0) - (x1 - x0) * (y3 - y0)) / b0
    b3 = ((x1 - x0) * (y2 - y0) - (x2 - x0) * (y1 - y0)) / b0 
    
    If b1>0 And b2>0 And b3>0 Then Return True
    
End Function

;~IDEal Editor Parameters:
;~C#Blitz3D198