;-------------------------------------------------------------------------------
; * Music fading library V2
;-------------------------------------------------------------------------------
; Version 2006 Oct 08 Rev 03
; Original: MGW\Code\Music Storage.BB
; Created by Teasy <TheChange@yahoo.com> (in case you didn't know! :o)
;-------------------------------------------------------------------------------
; Limitations: this system requires sounds to be looped.
; No error checking is performed when playing a new sound/music.
; Expansions:
; / add ability to play music directly from file
;   (only stream formats supported currently)
;   options:
;   X can add a separate (wrapped) function
;   / or added functionality to the existing,
;     but that would slow it down a lot
;     ! not entirely true (1 additional typecast)
;   for the PlayMusic command;
;   handling filenames rather than handles
; / added ability to allow invalid handles (skips handle)
; / added ability to error check before playing
; / add ability to play non-looped sounds
;   (a smart few extra ChannelPlaying checks)
; - then merge with Sound Lib to be able to play:
;   - thrust sound (looping) with quick fade-in/out on command (test case! :)
;     - using the same commands as normal sounds
;     - as well as music
;   - psb process outline:
;     - implement the playsample algorithm in this lib
;     - then change these command names for more uniformity (rather than just NewMusic, e.g. PlayAudio)
; - loadsoundprotected currently has ability to set default volume/pitch/panning etc
;   this conflicts with the current newmusic()
;   - one solution is moving the entire default volume/pitch/panning to the newmusic()
;   - another is using optional parameters, where, if unused, its value is not set (like in loadsoundprot)
;     - loadsoundprot can here also be modified to use globalvolume
;     - doc would include: if volume omitted in loadsoundprot, default vol is 1
;                          if volume omitted in newmusic(), volume from loadsoundprot is used
;-------------------------------------------------------------------------------

; If an audio handle refers to a non-existing audio object:
;   bAllowInvalidHandles = False: halt with an error (safe)
;   bAllowInvalidHandles = True: skip the object
Const Music_AllowInvalidHandles = False

; If a LoadSound handle or PlayMusic filename is invalid:
;   bPlayErrorCheck = False: continue regardless
;   bPlayErrorCheck = True: halt with an error (safe)
Const Music_PlayErrorCheck = False 

;@ - can make this 1000
;  - change name to: Audio_DefaultFadeTime
;  / add to name: default
Const Music_DefaultFadingTime = 2500  ; Millisecs

Const MusicStatus_Idle    = 0
Const MusicStatus_FadeIn  = 1
Const MusicStatus_FadeOut = 2

;-------------------------------------------------------------------------------

;@ psb new name: Audio_Channel
Type Music_Module
    Field Status      ; 1 = fade in, 2 = play, 3 = fade out
    Field Channel     ; channel of music/sound playing
    Field Started     ; timer handle
    Field Volume#     ; current/active volume (0..1)
    Field MaxVolume#  ; maximum volume for this music (0..1, default = 1.0)
    Field KeepAlive   ; keep this object after fade out (never delete)
    Field FadingTime  ; custom duration of fade
End Type

;-------------------------------------------------------------------------------

Global Music_GlobalVolume# = 1.0  ; Default: 1.0 = 100%

;-------------------------------------------------------------------------------

; Fade a new music in and fade those currently playing out
Function NextMusic( SoundHandle$ , MusicVolume#=1.0 )
; SoundHandle: either a pre-loaded sound (PlaySound) or a filename (PlayMusic)
; MusicVolume: optional maximum music volume (0..1)
    ; Fade out all currently playing
    FadeOutAllMusic
    ; Start playing a new one
    MusicHandle = NewMusic( SoundHandle )
    ; And fade it in
    FadeInMusic MusicHandle
    ; Trackability if desired
    Return MusicHandle
End Function

; Start fading a music in that's already playing
Function FadeInMusic( MusicHandle , FadingTime=Music_DefaultFadingTime )
; FadingTime: optional custom music fade-in duration
    p.Music_Module = Object.Music_Module MusicHandle
    If p = Null
        If Music_AllowInvalidHandles
            Return 
        Else 
            RuntimeError "FadeInMusic( MusicHandle )"+Chr(13)+"MusicHandle is not a valid handle"
        EndIf 
    EndIf 
    p\FadingTime = FadingTime
    If p\Volume < 1
        Select p\Status
            Case MusicStatus_Idle
                p\Started = MilliSecs
            Case MusicStatus_FadeIn
            Case MusicStatus_FadeOut
                p\Started = MilliSecs - p\Volume * p\FadingTime
        End Select
    EndIf
    p\Status = MusicStatus_FadeIn
End Function

; Start fading a music out (already playing)
Function FadeOutMusic( MusicHandle , FadingTime=Music_DefaultFadingTime )
; FadingTime: optional custom music fade-out duration
    p.Music_Module = Object.Music_Module MusicHandle
    If p = Null
        If Music_AllowInvalidHandles
            Return 
        Else 
            RuntimeError "FadeOutMusic( MusicHandle )"+Chr(13)+"MusicHandle is not a valid handle"
        EndIf 
    EndIf 
    p\FadingTime = FadingTime
    If p\Volume > 0
        Select p\Status
            Case MusicStatus_Idle
                p\Started = MilliSecs
            Case MusicStatus_FadeIn
                p\Started = MilliSecs - ( p\FadingTime - p\Volume * p\FadingTime )
            Case MusicStatus_FadeOut
        End Select
        p\Status = MusicStatus_FadeOut
    EndIf
End Function

; Fade all playing music out
Function FadeOutAllMusic()
    ; Check all music playing/fading
    For p.Music_Module = Each Music_Module
        ; If this one is not fading out or doesn't have its volume turned down already
        If p\Status <> MusicStatus_FadeOut And p\Volume > 0
            ; Fade it out
            FadeOutMusic Handle p
        EndIf
    Next
End Function

; Start playing a new sound/music
Function NewMusic( SoundHandle$ , MusicVolume#=1.0 , Silent=True , KeepAlive=False )
; MusicVolume: with optional maximum volume (0..1)
; Silent: and whether to play silently (True) or at full volume instantly (False)
; KeepAlive: music object is always accessible (never deleted)
    p.Music_Module = New Music_Module
    p\Volume = Not Silent
    p\MaxVolume = MusicVolume
    p\Status = MusicStatus_Idle
    ; SoundHandle is a number (a handle)
    If Int SoundHandle = SoundHandle
        If Music_PlayErrorCheck
            If SoundHandle = 0
                RuntimeError "NewMusic()->PlaySound"+Chr(13)+"Invalid SoundHandle"
            EndIf 
        EndIf 
        p\Channel = PlaySound( SoundHandle )
        If Music_PlayErrorCheck
            If p\Channel = 0
                RuntimeError "NewMusic()->PlaySound"+Chr(13)+"Sound not supported"
            EndIf 
        EndIf 
    ; SoundHandle is a string (a file)
    Else 
        If Music_PlayErrorCheck
            If FileType( SoundHandle ) = 1
                RuntimeError "NewMusic()->PlayMusic"+Chr(13)+"Track does not exist:"+Chr(13)+SoundHandle
            EndIf 
        EndIf 
        p\Channel = PlayMusic( SoundHandle )
        If Music_PlayErrorCheck
            If p\Channel = 0
                RuntimeError "NewMusic()->PlayMusic"+Chr(13)+"Track not supported:"+Chr(13)+SoundHandle
            EndIf 
        EndIf 
    EndIf 
    ChannelVolume p\Channel , p\Volume * p\MaxVolume * Music_GlobalVolume
    Return Handle p
End Function

; Change the keep-alive state; 0 = auto delete on fade out, 1 = never delete, -1 = toggle current state
Function StickMusic( MusicHandle , KeepAlive=True )
    p.Music_Module = Object.Music_Module MusicHandle
    If p = Null
        If Music_AllowInvalidHandles
            Return 
        Else 
            RuntimeError "SetMusicAlive( MusicHandle , KeepAlive )"+Chr(13)+"MusicHandle is not a valid handle"
        EndIf 
    EndIf 
    If KeepAlive = -1
        p\KeepAlive = Not p\KeepAlive
    Else
        p\KeepAlive = KeepAlive
    EndIf
End Function

; Eternal silence
Function StopMusic( MusicHandle )
    p.Music_Module = Object.Music_Module MusicHandle
    If p = Null
        If Music_AllowInvalidHandles
            Return 
        Else 
            RuntimeError "StopMusic( MusicHandle )"+Chr(13)+"MusicHandle is not a valid handle"
        EndIf 
    EndIf 
    If ChannelPlaying( p\Channel )
        StopChannel p\Channel
    EndIf 
    Delete p
End Function

; Protected music playing detection
Function MusicPlaying( MusicHandle )
    p.Music_Module = Object.Music_Module MusicHandle
    ; not playing if: invalid handle
    If p = Null Then Return False 
    ; not playing if: sound ended prematurely or entirely
    If Not ChannelPlaying( p\Channel ) Then Return False 
    ; not playing if: sound still playing, but volume zero
    If p\Volume = 0 Then Return False 
    ; otherwise IS playing
    Return True 
End Function

; Update all music fading (if fading at all)
Function UpdateMusic()
    Local TimeSinceStart#
    ; Check all music playing/fading
    For p.Music_Module = Each Music_Module
        ; Sound ended prematurely!
        If Not ChannelPlaying( p\Channel )
            ; This object must stay alive
            If p\KeepAlive
                ; Turn volume down and idle
                p\Volume = 0
                p\Status = MusicStatus_Idle
            ; Object may vanish
            Else 
                ; Eternal silence sets in
                StopChannel p\Channel
                Delete p
            EndIf 
        EndIf 
        Select p\Status
            ; No volume changes
            Case MusicStatus_Idle
                ; Except, if the channel suddenly becomes non-stick
                If p\KeepAlive = False 
                    ; And its volume is zero
                    If p\Volume = 0
                        ; Eternal silence sets in
                        If ChannelPlaying( p\Channel )
                            StopChannel p\Channel
                        EndIf 
                        Delete p
                    EndIf 
                EndIf 
            ; Volume up
            Case MusicStatus_FadeIn
                TimeSinceStart = MilliSecs - p\Started
                ; Duration exceeds fading time
                If TimeSinceStart >= p\FadingTime
                    ; Play at full volume
                    p\Volume = 1
                    ChannelVolume p\Channel , p\Volume * p\MaxVolume * Music_GlobalVolume
                    ; And switch to idle
                    p\Status = MusicStatus_Idle
                ; Not yet done fading
                Else
                    p\Volume = TimeSinceStart / p\FadingTime
                    ChannelVolume p\Channel , p\Volume * p\MaxVolume * Music_GlobalVolume
                EndIf
            ; Volume down
            Case MusicStatus_FadeOut
                TimeSinceStart = MilliSecs - p\Started
                ; Duration exceeds fading time
                If TimeSinceStart >= p\FadingTime
                    ; Handle management requested
                    If p\KeepAlive
                        ; Silent
                        p\Volume = 0
                        ChannelVolume p\Channel , 0
                        ; Idle
                        p\Status = MusicStatus_Idle
                    ; No handle management required
                    Else
                        ; Eternal silence sets in
                        If ChannelPlaying( p\Channel )
                            StopChannel p\Channel
                        EndIf 
                        Delete p
                    EndIf
                ; Not yet done fading
                Else
                    p\Volume = 1.0 - TimeSinceStart / p\FadingTime
                    ChannelVolume p\Channel , p\Volume * p\MaxVolume * Music_GlobalVolume
                EndIf
        End Select
    Next
End Function

; Set and/or get global music volume
Function MusicVolume( NewVolume#=-1 )
    If NewVolume >= 0 And NewVolume <= 1
        Music_GlobalVolume = NewVolume
    EndIf 
    Return Music_GlobalVolume
End Function 

