

;-------------------------------------------------------------------------------
;AppTitle "Multi-screen transition system"
;-------------------------------------------------------------------------------
; Created by Teasy <TheChange@yahoo.com>
; Version 1-E / 12 Aug 2007 @ 0123
; Updates:
; - 2011-12-02:
;   - added 'setpermanent' function to make image 'original' (keep after transition)
;   - made 'update' function return True when transition is complete
;-------------------------------------------------------------------------------
;Include "\Dev\Blitz\Lib\MaskFade 1a4.Lib.BB"
;Include "\Dev\Blitz\Lib\TimerSystem 4b.BB"
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Type screen
;-------------------------------------------------------------------------------
    Field x,y
    Field img
    Field original  ; if image is not original (referenced), do not wipe it on screen delete
    Field masked    ; draw image masked
;-------------------------------------------------------------------------------
End Type 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
; Transition types:
;-------------------------------------------------------------------------------
Const transitiontype_instantswitch      =  1
Const transitiontype_interruptedswitch  =  2
Const transitiontype_moveup             = 11
Const transitiontype_movedown           = 12
Const transitiontype_moveleft           = 13
Const transitiontype_moveright          = 14
Const transitiontype_instantfade        = 21
Const transitiontype_interruptedfade    = 22
;-------------------------------------------------------------------------------
;  1: Instant switch from active to new (default); same as transition_switch()
;  2: Instant black out active, wait, instant display to new
; 11: Moving up (active screen out (upwards), new screen in (upwards from below))
; 12: Moving down
; 13: Moving left
; 14: Moving right
; 21: Instant mask fade (fade out + fade in simultaneously)
; 22: Interrupted mask fade (fade out active, wait, fade in new)
;-------------------------------------------------------------------------------
Global transition_type      ; current transition type (if in progress)
Global transition_active    ; transition in progress
Global transition_duration  ; how long this transition takes (in ms)
Global transition_delay     ; how long to wait between transitions (if multiple stages)
;-------------------------------------------------------------------------------
Global transition_activescreen.screen   ; screen being displayed
Global transition_newscreen.screen      ; screen to be displayed
Global transition_started               ; time transition begun
Global transition_stage                 ; 0: old out, 1: wait, 2: new in
;-------------------------------------------------------------------------------
Global transition_width     ; dimensions of each screen
Global transition_height
Global transition_imagefront    ; transparent activescreen (foreground)
Global transition_imagefinal    ; background newscreen with transparent activescreen on top
Global transition_bufferfront   ; quick reference to image buffers
Global transition_bufferfinal
;-------------------------------------------------------------------------------
Global transition_initialized
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_init( w = -1 , h = -1 )
;-------------------------------------------------------------------------------
    ; clear system (if already initialized)
    transition_exit
    
    ; reset flags
    transition_type = 0
    transition_active = False 
    transition_duration = 0
    transition_delay = 0
    transition_activescreen = Null
    transition_newscreen = Null 
    transition_started = 0
    transition_stage = 0

    ; specific dimensions, or from current resolution
    If w = -1
        w = GraphicsWidth()
    EndIf 
    If h = -1 
        h = GraphicsHeight()
    EndIf 
    transition_width = w
    transition_height = h

    ; mask filter fade system
    MaskFade_Init
    transition_imagefront = CreateImage( w,h )
    transition_imagefinal = CreateImage( w,h )
    transition_bufferfront = ImageBuffer( transition_imagefront )
    transition_bufferfinal = ImageBuffer( transition_imagefinal )

    ; all done!
    transition_initialized = True 
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_exit()
;-------------------------------------------------------------------------------
    If transition_initialized
        FreeImage transition_imagefront
        FreeImage transition_imagefinal
        MaskFade_Exit 
        transition_initialized = False 
        Delete Each screen
    EndIf 
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_newscreen.screen( activate = True , img = 0 )
;-------------------------------------------------------------------------------
    Local p.screen

    ; screen object
    p = New screen
    p\x = 0
    p\y = 0
    p\masked = False 

    ; create new image or passed existing image as parameter
    If img = 0
        p\img = CreateImage( transition_width,transition_height )
        p\original = True 
    Else 
        p\img = img
        p\original = False 
    EndIf 

    ; instantly activate (become new screen)
    If activate
        transition_setactive p
    EndIf 
    
    Return p
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_deletescreen( p.screen )
;-------------------------------------------------------------------------------
    ; do not delete image if it was not original
    If p\original
        FreeImage p\img
    EndIf 
    
    Delete p
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_switch( newscreen.screen )
;-------------------------------------------------------------------------------
    If newscreen <> Null 
        transition_start newscreen
        transition_update 
    EndIf 
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_setactive( newscreen.screen )
;-------------------------------------------------------------------------------
    If newscreen <> Null 
        ; instantly switch to new screen (if not already)
        transition_activescreen = newscreen

        ; forget current transition (if any)
        transition_active = False 
        transition_newscreen = Null 

        ; reset transition specific parameters
        transition_activescreen\x = 0
        transition_activescreen\y = 0
    EndIf 
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_setmask( p.screen , maskred = -1 , maskgreen = -1 , maskblue = -1 )
;-------------------------------------------------------------------------------
    ; new mask specified
    If maskred <> -1 Or maskgreen <> -1 Or maskblue <> -1
        ; set mask and remember to use it
        MaskImage p\img , maskred,maskgreen,maskblue
        p\masked = True 
    ; nothing specified
    Else 
        ; don't use mask
        p\masked = False 
    EndIf 
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_setpermanent( p.screen , permanent=True )
;-------------------------------------------------------------------------------
    If p <> Null
        p\original = permanent
    End If
;-------------------------------------------------------------------------------
End Function
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_start( newscreen.screen , transition = transitiontype_instantswitch , duration = 1000 , waitdelay = 500 )
;-------------------------------------------------------------------------------
;   Param newscreen.screen: screen to transit to
;   Param transition: type of transition to invoke
;   Param duration: duration of total motion or individual fade duration
;   Param waitdelay: delay between interrupted switch or fade
;-------------------------------------------------------------------------------
; Examples:
;   transition  2: instant blank, wait 500 ms, instant appear
;   transition 11: both move in 1000 ms
;   transition 21: both fade in 1000 ms
;   transition 22: fade out 1000 ms, wait 500 ms, fade in 1000 ms
;-------------------------------------------------------------------------------
    ; if in a transition at the moment
    If transition_active
        ; reset current
        transition_setactive transition_activescreen
        ; then instantly switch to (old) new screen first (and reset it)
        transition_setactive transition_newscreen
    EndIf 

    ; commence new transition
    transition_newscreen = newscreen
    transition_type = transition
    transition_duration = duration
    transition_delay = waitdelay
    transition_active = True 
    transition_stage = 0
    transition_started = MilliSecs

    ; set transition specific parameters
    Select transition_type
        Case transitiontype_instantswitch
        Case transitiontype_interruptedswitch
        Case transitiontype_moveup      ; active moves up, new moves up from below
            ; place newscreen below current
            transition_newscreen\x = 0
            transition_newscreen\y = transition_height
        Case transitiontype_movedown    ; active moves down, new moves down from above
            ; place newscreen above current
            transition_newscreen\x = 0
            transition_newscreen\y = -transition_height
        Case transitiontype_moveleft    ; active moves left, new moves left from right
            ; place newscreen right of current
            transition_newscreen\x = transition_width
            transition_newscreen\y = 0
        Case transitiontype_moveright   ; active moves right, new moves right from left
            ; place newscreen left of current
            transition_newscreen\x = -transition_width
            transition_newscreen\y = 0
        Case transitiontype_instantfade
            ; set transparent first (activescreen is visible)
            MaskFade_Set +1
            ; set time to fade and initiate
            MaskFade_Duration transition_duration
            MaskFade_Out
        Case transitiontype_interruptedfade
            ; set transparent first; +1 (visible) -> -1 (invisible)
            MaskFade_Set +1
            ; set time to fade and initiate
            MaskFade_Duration transition_duration
            MaskFade_Out
    End Select 
;-------------------------------------------------------------------------------
End Function
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_update()
;-------------------------------------------------------------------------------
    Local difference
    Local done = False
    
    If transition_active
        Select transition_type
        ;-----------------------------------------------------------------------
            Case transitiontype_instantswitch
            ;-------------------------------------------------------------------
                ; already done!
                transition_setactive transition_newscreen
                done = True
        ;-----------------------------------------------------------------------
            Case transitiontype_interruptedswitch
            ;-------------------------------------------------------------------
                ; which stage
                Select transition_stage
                ;---------------------------------------------------------------
                    Case 0  ; first stage
                    ;-----------------------------------------------------------
                        ; first stage complete -> second stage
                        transition_stage = 1
                        transition_started = MilliSecs
                        ; not showing anything right now
                        transition_activescreen = Null 
                ;---------------------------------------------------------------
                    Case 1  ; center stage
                    ;-----------------------------------------------------------
                        ; duration of second stage has elapsed
                        If MilliSecs - transition_started >= transition_delay
                            ; second stage complete -> last stage
                            transition_stage = 2
                            ; now showing new screen
                            transition_activescreen = transition_newscreen
                        EndIf 
                ;---------------------------------------------------------------
                    Case 2  ; final stage
                    ;-----------------------------------------------------------
                        ; last stage complete -> reset transition
                        transition_setactive transition_activescreen
                        done = True
                ;---------------------------------------------------------------
                End Select 
        ;-----------------------------------------------------------------------
            Case transitiontype_moveup
            ;-------------------------------------------------------------------
                ; measure time
                difference = MilliSecs - transition_started
                ; overdue
                If difference >= transition_duration
                    ; transition complete
                    transition_setactive transition_newscreen
                    done = True
                ; not yet done
                Else 
                    ; set position of activescreen relative to time (scale+offset)
                    transition_activescreen\y = 0 - difference * transition_height / transition_duration
                    ; then set newscreen relative to activescreen (offset)
                    transition_newscreen\y = transition_activescreen\y + transition_height
                EndIf 
        ;-----------------------------------------------------------------------
            Case transitiontype_movedown
            ;-------------------------------------------------------------------
                ; measure time
                difference = MilliSecs - transition_started
                ; overdue
                If difference >= transition_duration
                    ; transition complete
                    transition_setactive transition_newscreen
                    done = True
                ; not yet done
                Else 
                    ; set position of activescreen relative to time (scale+offset)
                    transition_activescreen\y = difference * transition_height / transition_duration
                    ; then set newscreen relative to activescreen (offset)
                    transition_newscreen\y = transition_activescreen\y - transition_height
                EndIf 
        ;-----------------------------------------------------------------------
            Case transitiontype_moveleft
            ;-------------------------------------------------------------------
                ; measure time
                difference = MilliSecs - transition_started
                ; overdue
                If difference >= transition_duration
                    ; transition complete
                    transition_setactive transition_newscreen
                    done = True
                ; not yet done
                Else 
                    ; set position of activescreen relative to time (scale+offset)
                    transition_activescreen\x = 0 - difference * transition_width / transition_duration
                    ; then set newscreen relative to activescreen (offset)
                    transition_newscreen\x = transition_activescreen\x + transition_width
                EndIf 
        ;-----------------------------------------------------------------------
            Case transitiontype_moveright
            ;-------------------------------------------------------------------
                ; measure time
                difference = MilliSecs - transition_started
                ; overdue
                If difference >= transition_duration
                    ; transition complete
                    transition_setactive transition_newscreen
                    done = True
                ; not yet done
                Else 
                    ; set position of activescreen relative to time (scale+offset)
                    transition_activescreen\x = difference * transition_width / transition_duration
                    ; then set newscreen relative to activescreen (offset)
                    transition_newscreen\x = transition_activescreen\x - transition_width
                EndIf 
        ;-----------------------------------------------------------------------
            Case transitiontype_instantfade
            ;-------------------------------------------------------------------
                ; measure time
                difference = MilliSecs - transition_started
                ; overdue
                If difference >= transition_duration
                    ; transition complete
                    transition_setactive transition_newscreen
                    done = True
                    ; reset mask fade
                    MaskFade_Set -1
                ; not yet done
                Else 
                    MaskFade_Update
                EndIf 
        ;-----------------------------------------------------------------------
            Case transitiontype_interruptedfade
            ;-------------------------------------------------------------------
                ; which stage
                Select transition_stage
                ;---------------------------------------------------------------
                    Case 0  ; first stage (fade to black)
                    ;-----------------------------------------------------------
                        ; measure time
                        difference = MilliSecs - transition_started
                        ; overdue
                        If difference >= transition_duration
                            ; reset mask fade (transparent)
                            MaskFade_Set -1
                            ; first stage complete -> second stage
                            transition_stage = 1
                            transition_started = MilliSecs
                            ; not showing anything right now
                            transition_activescreen = Null 
                        ; not yet done
                        Else 
                            MaskFade_Update
                        EndIf 
                ;---------------------------------------------------------------
                    Case 1  ; center stage (blackness)
                    ;-----------------------------------------------------------
                        ; duration of second stage has elapsed
                        If MilliSecs - transition_started >= transition_delay
                            ; second stage complete -> last stage
                            transition_stage = 2
                            transition_started = MilliSecs 
                            ; set mask to cover newscreen (black)
                            MaskFade_Set -1
                            ; then start fading out
                            MaskFade_In
                        EndIf 
                ;---------------------------------------------------------------
                    Case 2  ; final stage (fade new in)
                    ;-----------------------------------------------------------
                        ; measure time
                        difference = MilliSecs - transition_started
                        ; overdue
                        If difference >= transition_duration
                            ; last stage complete -> reset transition
                            transition_setactive transition_newscreen
                            done = True
                            ; reset mask fade (transparent)
                            MaskFade_Set -1
                        ; not yet done
                        Else 
                            MaskFade_Update
                        EndIf 
                ;---------------------------------------------------------------
                End Select 
        ;-----------------------------------------------------------------------
        End Select
    End If
    
    Return done
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_display()
;-------------------------------------------------------------------------------
    Local drawtype  ; 0: none, 1: active, 2: active & new
    Local drawwhich.screen
    Local buffer

    ; transition is active
    If transition_active
        Select transition_type
        ;-----------------------------------------------------------------------
            Case transitiontype_instantswitch
            ;-------------------------------------------------------------------
                ; just in case this point is ever reached
                drawtype = 1
                drawwhich = transition_activescreen
        ;-----------------------------------------------------------------------
            Case transitiontype_interruptedswitch
            ;-------------------------------------------------------------------
                ; always drawing active screen (changed in update)
                drawtype = 1
                drawwhich = transition_activescreen
        ;-----------------------------------------------------------------------
            Case transitiontype_moveup
            ;-------------------------------------------------------------------
                ; draw both active and new
                drawtype = 2
        ;-----------------------------------------------------------------------
            Case transitiontype_movedown
            ;-------------------------------------------------------------------
                ; draw both active and new
                drawtype = 2
        ;-----------------------------------------------------------------------
            Case transitiontype_moveleft
            ;-------------------------------------------------------------------
                ; draw both active and new
                drawtype = 2
        ;-----------------------------------------------------------------------
            Case transitiontype_moveright
            ;-------------------------------------------------------------------
                ; draw both active and new
                drawtype = 2
        ;-----------------------------------------------------------------------
            Case transitiontype_instantfade
            ;-------------------------------------------------------------------
                ; remember current buffer
                buffer = GraphicsBuffer()
                
                ; make activescreen transparent (on top)
                SetBuffer transition_bufferfront
                If transition_activescreen <> Null 
                    If transition_activescreen\masked
                        DrawImage transition_activescreen\img , 0,0
                    Else 
                        DrawBlock transition_activescreen\img , 0,0
                    EndIf 
                EndIf 
                MaskFade_Draw
                
                ; make transparent activescreen overlap newscreen (background)
                SetBuffer transition_bufferfinal
                If transition_newscreen <> Null 
                    If transition_newscreen\masked
                        DrawImage transition_newscreen\img , 0,0
                    Else 
                        DrawBlock transition_newscreen\img , 0,0
                    EndIf 
                EndIf 
                DrawImage transition_imagefront , 0,0

                ; switch back to current buffer and draw result
                SetBuffer buffer
                DrawImage transition_imagefinal , 0,0
                
                ; don't draw anything else
                drawtype = 0
        ;-----------------------------------------------------------------------
            Case transitiontype_interruptedfade
            ;-------------------------------------------------------------------
                Select transition_stage
                ;---------------------------------------------------------------
                    Case 0  ; fade to black
                    ;-----------------------------------------------------------
                        ; draw activescreen
                        If transition_activescreen <> Null 
                            If transition_activescreen\masked
                                DrawImage transition_activescreen\img , 0,0
                            Else 
                                DrawBlock transition_activescreen\img , 0,0
                            EndIf 
                        EndIf 

                        ; then overlay it with mask
                        MaskFade_Draw

                        ; don't draw anything else
                        drawtype = 0
                ;---------------------------------------------------------------
                    Case 1  ; total blackness
                    ;-----------------------------------------------------------
                        ; don't draw anything at all
                        drawtype = 0
                ;---------------------------------------------------------------
                    Case 2  ; fade new in
                    ;-----------------------------------------------------------
                        ; draw newscreen
                        If transition_newscreen <> Null 
                            If transition_newscreen\masked
                                DrawImage transition_newscreen\img , 0,0
                            Else 
                                DrawBlock transition_newscreen\img , 0,0
                            EndIf 
                        EndIf 

                        ; then overlay it with mask
                        MaskFade_Draw

                        ; don't draw anything else
                        drawtype = 0
                ;---------------------------------------------------------------
                End Select 
        ;-----------------------------------------------------------------------
        End Select 
    Else 
        ; transition is not active: draw only active screen
        drawtype = 1
        drawwhich = transition_activescreen
    EndIf 

    Select drawtype
        Case 0  ; don't draw
        Case 1  ; draw only active
            ; only draw if there is something to draw
            If drawwhich <> Null 
                ; image should be drawn with mask
                If drawwhich\masked
                    DrawImage drawwhich\img , drawwhich\x,drawwhich\y
                ; draw without mask (default)
                Else 
                    DrawBlock drawwhich\img , drawwhich\x,drawwhich\y
                EndIf 
            EndIf 
        Case 2  ; draw both (new overlaps active, if that ever happens)
            If transition_activescreen <> Null 
                If transition_activescreen\masked
                    DrawImage transition_activescreen\img , transition_activescreen\x,transition_activescreen\y
                Else 
                    DrawBlock transition_activescreen\img , transition_activescreen\x,transition_activescreen\y
                EndIf 
            EndIf 
            If transition_newscreen <> Null 
                If transition_newscreen\masked
                    DrawImage transition_newscreen\img , transition_newscreen\x,transition_newscreen\y
                Else 
                    DrawBlock transition_newscreen\img , transition_newscreen\x,transition_newscreen\y
                EndIf 
            EndIf 
    End Select 
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
Function transition_demo1( sx = 400 , sy = 300 )
;-------------------------------------------------------------------------------
    Local BackBuffer 
    Local p1.screen
    Local p2.screen
    Local p1b,p2b
    Local n
    Local transition

    Graphics sx,sy , 0,3
    transition_init

    ; active screen
    p1 = transition_newscreen()
    ; new screen (currently invisible)
    p2 = transition_newscreen( False )

    ; store quick buffer references
    BackBuffer = BackBuffer()
    p1b = ImageBuffer( p1\img )
    p2b = ImageBuffer( p2\img )

    ; initial transition type
    transition = transitiontype_instantfade

    Repeat 
        ; update time
        MilliSecs = MilliSecs()
        ; update input
        For n = 2 To 2+8
            If KeyHit( n )  ; 1..8 keys -> select transition type
                Select n-1
                    Case 1  transition = transitiontype_instantswitch
                    Case 2  transition = transitiontype_interruptedswitch
                    Case 3  transition = transitiontype_moveup
                    Case 4  transition = transitiontype_movedown
                    Case 5  transition = transitiontype_moveleft
                    Case 6  transition = transitiontype_moveright
                    Case 7  transition = transitiontype_instantfade
                    Case 8  transition = transitiontype_interruptedfade
                End Select 
            EndIf 
        Next 
        If KeyHit( 57 )  ; space -> commence transition
            ; if already in a transition
            If transition_active
                ; quickly jump to end of transition
                If transition_newscreen = p1
                    transition_switch p1
                ElseIf transition_newscreen = p2
                    transition_switch p2
                EndIf 
            EndIf 
            ; followed by the new transition
            If transition_activescreen = p1
                transition_start p2 , transition
            ElseIf transition_activescreen = p2
                transition_start p1 , transition
            EndIf 
        EndIf 
        ; update logic
        transition_update
        ; screen 1: blue rects
        SetBuffer p1b
        For n = 1 To 10
            Color 0,0,Rand( 0,255 )
            Rect Rand( -sx/4,sx/1.5 ),Rand( -sy/4,sy/1.5 ) , Rand( 0,sx ),Rand( 0,sy )
        Next 
        Color 255,255,255
        Text sx/2,sy/2 , "1" , True,True
        ; screen 2: red ovals
        SetBuffer p2b
        For n = 1 To 10
            Color Rand( 0,255 ),0,0
            Oval Rand( -sx/4,sx/1.5 ),Rand( -sy/4,sy/1.5 ) , Rand( 0,sx ),Rand( 0,sy )
        Next 
        Color 255,255,255
        Text sx/2,sy/2 , "2" , True,True
        ; what you see
        SetBuffer BackBuffer
        Cls 
        transition_display
        If Not KeyDown( 28 )
            Color 192,224,255
            Text 0,  0 , "Type: "+transition+" ("+transition_type+")   Active: "+transition_active+"   Stage: "+transition_stage+"   Time: "+(MilliSecs-transition_started)/1000
            Text 0, 20 , "Select transition type:"
            Text 0, 30 , "<1> Instant switch"
            Text 0, 40 , "<2> Interrupted switch (500ms)"
            Text 0, 50 , "<3> Move up (1000ms)"
            Text 0, 60 , "<4> Move down (1000ms)"
            Text 0, 70 , "<5> Move left (1000ms)"
            Text 0, 80 , "<6> Move right (1000ms)"
            Text 0, 90 , "<7> Instant fade (1000ms)"
            Text 0,100 , "<8> Interrupted fade (1000ms, 500ms)"
            Text 0,120 , "<Space> Start transition"
            Text 0,130 , "<Enter> Hide text"
            If transition_activescreen <> Null 
                Text 0,150 , "Active ("+transition_activescreen\x+","+transition_activescreen\y+")"
            EndIf 
            If transition_newscreen <> Null 
                Text 0,160 , "New ("+transition_newscreen\x+","+transition_newscreen\y+")"
            EndIf 
        EndIf 
        Flip 
    Until KeyHit( 1 )
    End
;-------------------------------------------------------------------------------
End Function 
;-------------------------------------------------------------------------------


;-------------------------------------------------------------------------------
;transition_demo1
;-------------------------------------------------------------------------------


;~IDEal Editor Parameters:
;~C#Blitz3D1105