;-------------------------------------------------------------------------------
; * Protected sample player
; Version 2010 Mar 01
;-------------------------------------------------------------------------------
;
;   PlaySample Sample , TimeOut
;
; This function can prevent the same Sample from being played simultaneously.
; You can specify the minimum overlap (TimeOut) in milliseconds, where 0 ms is
; simultaneously.  When playing the same Sample after the TimeOut, a new sound
; will overlap the already playing sound.  Every call to PlaySample
; automatically updates all channels.
; Returns False when the sound to be played overlaps; no new sound is played.
; Otherwise returns the handle of the newly playing sound, like PlaySound().
;
;   UpdateChannels
;
; You can manually clean up idle channels using UpdateChannels, even though
; this is also done in PlaySample, and not really necessary.
;
;   SoundOff
;
; You can force a total sound off using SoundOff.
;
;   LoadSoundProtected(
;       sFileName$,
;       Looping=False,
;       DefaultVolume#=-1,
;       DefaultPanning#=-2,
;       DefaultPitch#=-1
;   )
;
; Safely validate and load a sound (with optional defaults).
; Causes an explicit RuntimeError on any errors during load.
;-------------------------------------------------------------------------------
; changelog:
; 01 mar 2010: added test function and debugging values to playsample
;-------------------------------------------------------------------------------
; future expansions:
; - fade in / out on channels
;   - requires updatechannels to be called constantly
;   ! implemented in Music Fade Lib (see there for more)
; - music support (like MGW)
;   - with fade in / out on channels
;   ! implemented in Music Fade Lib (see there for more)
; - merge with Music Fade Lib when new features are implemented there
;   - but make sure this file is
;     1) backward compatible
;     2) still just as easy to use when merged
;-------------------------------------------------------------------------------

Type ActiveChannel
    Field ChannelHandle
    Field SoundHandle
    Field Created
    Field TimeOut   ; How long to wait before being able to play same sound.
End Type

;-------------------------------------------------------------------------------

Function PlaySample ( SoundHandle , TimeOut = False )
    Local FindChannel.ActiveChannel
    Local FoundChannel.ActiveChannel
    Local Playing = False
    Local PlayNew = False
    
    For FindChannel = Each ActiveChannel
        If ChannelPlaying ( FindChannel\ChannelHandle )
            If FindChannel\SoundHandle = SoundHandle
                Playing = True
                FoundChannel = FindChannel
            End If
        Else
            Delete FindChannel
        End If
    Next
    
    If Playing
        If MilliSecs () - FoundChannel\Created >= FoundChannel\TimeOut
            PlayNew = True
            ;DebugLog "Old playing -> elapsed ("+FoundChannel\TimeOut+") -> New ("+TimeOut+")"
        ;Else
            ;DebugLog "Old playing -> not elapsed ("+FoundChannel\TimeOut+")"
        End If
    Else
        ;DebugLog "Not old playing -> new ("+TimeOut+")"
        PlayNew = True
    End If
    
    If PlayNew
        FoundChannel = New ActiveChannel
        FoundChannel\Created = MilliSecs ()
        FoundChannel\TimeOut = TimeOut
        FoundChannel\SoundHandle = SoundHandle
        FoundChannel\ChannelHandle = PlaySound ( SoundHandle )
        Return FoundChannel\ChannelHandle
    End If
End Function

Function UpdateChannels ()
    Local Channel.ActiveChannel
    For Channel = Each ActiveChannel
        If Not ChannelPlaying ( Channel\ChannelHandle ) Then Delete Channel
    Next
End Function

Function SoundOff ()
    Local Channel.ActiveChannel
    For Channel = Each ActiveChannel
        StopChannel Channel\ChannelHandle
        Delete Channel
    Next
End Function

; Validate and load a sound (with optional defaults)
Function LoadSoundProtected( sFileName$ , Looping=False,DefaultVolume#=-1,DefaultPanning#=-2,DefaultPitch#=-1 )
    Local hSound

    If FileType( sFileName ) <> 1
        RuntimeError "LoadSound() - File not found:"+Chr(13)+sFileName
    EndIf
    hSound = LoadSound( sFileName )
    If hSound = 0
        RuntimeError "LoadSound() - Unable to load sample:"+Chr(13)+sFileName
    EndIf

    If Looping
        LoopSound hSound
    EndIf
    If DefaultVolume >= 0 And DefaultVolume <= 1
        SoundVolume hSound , DefaultVolume
    EndIf
    If DefaultPanning >= -1 And DefaultPanning <= 1
        SoundPan hSound , DefaultPanning
    EndIf
    If DefaultPitch > -1
        SoundPitch hSound , DefaultPitch
    EndIf

    Return hSound
End Function

; Validate and load a sound (with optional defaults)
; - also play the sound quietly to see if it's *really* playable
Function LoadSoundProtectedFailSafe( sFileName$ , Looping=False,DefaultVolume#=-1,DefaultPanning#=-2,DefaultPitch#=-1 )
    Local hSound
    Local hChannel
    
    hSound = LoadSoundProtected( sFileName , Looping,DefaultVolume,DefaultPanning,DefaultPitch )
    hChannel = PlaySound( hSound )
    If Not ChannelPlaying( hChannel )
        RuntimeError "LoadSound() - Unable to play sample:"+Chr(13)+sFileName
    End If
    StopChannel hChannel
    
    Return hSound
End Function

Function PlaySample_Test()
    Local sound = LoadSound( "Q:\Dev\Blitz\Game\Futurity\Construct\2010\samples\beep4.wav" )
    Repeat
        If KeyDown( 57 )
            PlaySample sound , 1000
        End If
        Delay 1
    Until KeyHit(1)
    End
End Function


;~IDEal Editor Parameters:
;~C#Blitz3D198