; implementable features:
; option to use time object per speed object (rather than global for all speeds)

;-------------------------------------------------------------------------------
; * Time System
; Version 1.0 / 03 Jun 2007 / 2337
; Created by Teasy <TheChange@yahoo.com>
;-------------------------------------------------------------------------------
;
; This versatile library can be used as a generic timing system,
; including relative time and speed tracking.
; Being basically an advanced MilliSecs() replacement / wrapper
; with the following additional abilities:
; - change the speed of time.
; - create any number of time objects.
; - pause time.
; - measure in speed (units per interval)
;   or in time (like delta time).
;
; Examples of usage:
; - Moving an object at identical speeds on any computer.
;   See TimeSystem_DemoSimple and TimeSystem_DemoAdvanced.
; - Pausing time / time suspension.
;   See TimeSystem_DemoPause.
; - Pausing multiple objects moving at unique speeds.
;   See TimeSystem_DemoMulti.
;
; Notes:
; - This version does not take time wrapping into account.
;   Any time differences over ~24.5 days will result in time scales
;   falling off the chart, leading to ~24.5 days of time suspension,
;   which must be handled by the host program or the programmer :-)
; - Follow these guidelines to prevent time-wrapping:
;   - Call TimeSystem_Update more often than ~24.5 days.
;   - Don't pause a time object for more than ~24.5 days.
;   - Don't adjust a time object by more than ~24.5 days.
;   You can also force a time-wrap just to see what happens.
;
;-------------------------------------------------------------------------------
;
; Usage:
;
; - Basic
;
;    - UpdateTime()
;
;       - Update current time, speed objects and any active time objects.
;       - Can be run at any interval (below ~24.5 days).
;
;    - MilliSecs
;
;       - Current system time.
;       - Updated using TimeSystem_Update.
;
; - Speed tracking
;
;    - NewSpeed( time frame = 1 )
;
;       - Create a new speed object with specified number of milliseconds.
;       - Returns a SPEED object, which can be used directly (see below).
;       - Use a time frame of 1 ms for maximum precision (default).
;
;    - DeleteSpeed( speed object )
;
;       - Free a speed object safely.
;       - Pass parameter Null to free all speed objects.
;
;    - SpeedObject.SPEED \ Units
;
;       - Amount of speed units since last update.
;       - For example if 1 second passed since last update,
;         and current speed is 1000 milliseconds, the result will be 1.
;       - Example usage: pos = pos + speed * object\Units
;
;    - ChangeSpeed( speed object , time frame )
;
;       - Alter speed object's time frame,
;         which can be changed at any time.
;
; - Time tracking
;
;    - NewTime( time difference = 0 )
;
;       - Create a new time object with offset relative to system time.
;       - Returns a TIME object, which can be used directly (see below).
;       - Time difference is measured in milliseconds
;         and can be both in the past (-n) or in the future (+n).
;
;    - DeleteTime( time object )
;
;       - Safely free a time object.
;       - Specify a Null parameter to clear all time objects.
;       - If this object is used for speed tracking,
;         time system will revert back to system time.
;
;    - TimeObject.TIME \ Now
;
;       - This is a time object's current time.
;       - Relies completely on system time,
;         so TimeSystem_Update must be used as well.
;
;    - PauseTime( time object )
;
;       - Suspend a time object.
;
;    - ResumeTime( time object )
;
;       - Unpause a time object.
;
;    - ToggleTime( time object )
;
;       - Automatically switch between active and inactive.
;         Identical to PauseTime and ResumeTime.
;
;    - TimeObject.TIME \ Paused
;
;       - True if the time tracker is currently paused.
;
;    - SetSpeedTime( time object )
;
;       - Use time object for speed tracking,
;         to be able to use pause and resume.
;       - Use parameter Null to return to system time.
;
;    - ShiftTime( time object , time difference )
;
;       - Shift time object in time by specified number of milliseconds.
;       - Time can be shifted both in the past (-n) and in the future (+n).
;       - Specify a Null object to shift all time objects.
;
;-------------------------------------------------------------------------------


; current system time (absolute)
Global MilliSecs

; time system access
Global TimeSystem_.TimeSystem_


;-------------------------------------------------------------------------------


; time system base object
Type TimeSystem_

    ; last call to timesystem_update
    Field LastTime

    ; focused time object (as current time)
    Field SpeedTime.TIME

End Type


; time tracker
Type TIME

    ; whether timer is paused or running
    Field Paused

    ; current time (relatively adjusted)
    Field Now

    ; remember point of previous time adjustment
    ; (similar to LastTime in time system object)
    Field LastTime

End Type


; speed tracker
Type SPEED

    ; number of time units for measurement (ms)
    Field TimeFrame

    ; number of time frames elapsed since last update
    Field Units#

End Type


;-------------------------------------------------------------------------------


; initialize time system
Function TimeSystem_Init()

    TimeSystem_ = New TimeSystem_

    ; sync last update with current time
    UpdateTime True

End Function


; update all components of the time system
Function UpdateTime( resync = False )
    Local elapsed#
    Local time.TIME
    Local speed.SPEED

    ; current time
    MilliSecs = MilliSecs()

    ; forget difference if request to sync
    If Not resync
        ; time since last call
        elapsed = MilliSecs - TimeSystem_\LastTime
    EndIf

    ; update time objects
    For time = Each TIME
        If Not time\Paused
            time\LastTime = time\Now
            time\Now = time\Now + elapsed
        EndIf
    Next

    ; if using custom time for speed tracking
    If TimeSystem_\SpeedTime <> Null
        ; use it for speed measurement
        elapsed = TimeSystem_\SpeedTime\LastTime - TimeSystem_\SpeedTime\Now
    ; no custom time for speed tracking
    Else
        ; keep system time difference
    EndIf

    ; update speed objects
    For speed = Each SPEED
        ; spread time distance over current speed distance
        ; and convert distance to speed
        speed\Units = elapsed / speed\TimeFrame
    Next

    ; sync time
    TimeSystem_\LastTime = MilliSecs

End Function


;-------------------------------------------------------------------------------


; create a time tracker with specified offset to system time
Function NewTime.TIME( difference = 0 )
    Local p.TIME

    p = New TIME
    p\Paused = False

    ; offset from system time
    p\Now = MilliSecs + difference

    ; sync time difference update
    p\LastTime = p\Now

    Return p
End Function


; clear one or all time trackers
Function DeleteTime( p.TIME )

    If p <> Null
        Delete p
    Else
        Delete Each TIME
    EndIf

End Function


; suspend a time tracker
Function PauseTime( p.TIME )

    If Not p\Paused
        ; sync time difference update
        p\LastTime = p\Now

        ; invoke pause state
        p\Paused = True
    EndIf 

End Function


; unpause a time tracker
Function ResumeTime( p.TIME )

    If p\Paused
        ; exit pause state
        p\Paused = False
    EndIf 

End Function


; Toggle time tracker on/off
Function ToggleTime( p.TIME )
    
    If p\Paused 
        ResumeTime p
    Else 
        PauseTime p
    EndIf 
    
End Function


; shift one or all time objects in time by specified number of milliseconds
Function ShiftTime( p.TIME , difference = 0 )

    If p <> Null
        p\Now = p\Now + difference
    Else
        For p = Each TIME
            p\Now = p\Now + difference
        Next
    EndIf

End Function


; use a time object for speed measurements
Function SetSpeedTime( p.TIME )

    TimeSystem_\SpeedTime = p

End Function


;-------------------------------------------------------------------------------


; create a speed tracker with specified number of milliseconds
Function NewSpeed.SPEED( timeframe = 1 )
    Local p.SPEED

    p = New SPEED

    ; set number of time units for measurement
    p\TimeFrame = timeframe

    Return p
End Function


; clear one or all speed trackers
Function DeleteSpeed( p.SPEED )

    If p <> Null
        Delete p
    Else
        Delete Each SPEED
    EndIf

End Function


; alter speed object's time units
Function ChangeSpeed( p.SPEED , timeframe )

    p\TimeFrame = timeframe

End Function


;-------------------------------------------------------------------------------
;   Demos
;-------------------------------------------------------------------------------


Function TimeSystem_DemoAdvanced()
    Local w = 320
    Local h = 240

    Graphics w,h , 0,3
    SetFont LoadFont( "Verdana" , 13 , True )

    Local radx = w/2 * .75  ; marker rotation radius (oval)
    Local rady = h/2 * .75
    Local angle# = 0     ; marker angle
    Local vsync = True
    Local tsync = False
    Local timer = CreateTimer( 10 )
    Local bbw = w  ; background block width/height
    Local bbh = h
    Local bw = w / 10.  ; block width/height
    Local bh = h / 10.
    Local backimg = CreateImage( bbw,bbh )
    MidHandle backimg
    Local markimg = CreateImage( bw,bh )
    MidHandle markimg

    Local pixw = 1   ; pixel width/height
    Local pixh = 1
    Local stepw = 2  ; step width/height
    Local steph = 2
    Local square = False  ; square or round marker

    Local x,y    ; marker position
    Local ox,oy  ; draw offset
    Local c      ; color
    Local ms     ; milliseconds (current time)
    Local bx,by  ; background position
    Local switch ; toggles on/off each frame
    Local render ; render marker pixel (square/round)

    Local fps          ; number of frames rendered in last second
    Local frames       ; count number of frames during this second
    Local lastsecond   ; time since last second

    Local speed.SPEED = NewSpeed( 10 )  ; use 10 ms as time frame
    SeedRnd MilliSecs

    ; pre-render background
    .reset
    SetBuffer ImageBuffer( backimg )
    Cls
    ox = 0
    Repeat
        oy = 0
        Repeat
            c = Rand( 0,127 )
            Color c,c,c
            Rect ox,oy , pixw,pixh
            oy = oy + steph
        Until oy >= bbh
        ox = ox + stepw
    Until ox >= bbw
    ; clear marker
    SetBuffer ImageBuffer( markimg )
    Cls

    Repeat
        ; input
        If KeyHit( 15 )  ; tab
            If stepw = 2
                stepw = 1
                steph = 1
            Else
                stepw = 2
                steph = 2
            EndIf
            Goto reset
        EndIf
        If KeyHit( 28 )  ; enter
            tsync = Not tsync
        EndIf
        If KeyHit( 57 )  ; space
            vsync = Not vsync
        EndIf
        If KeyHit( 42 ) Or KeyHit( 54 )  ; shift
            square = Not square
            Goto reset
        EndIf
        ; timing
        UpdateTime
        ms = MilliSecs
        frames = frames + 1
        If ms - lastsecond >= 1000
            lastsecond = ms
            fps = frames
            frames = 0
        EndIf
        ; marker angle
        angle = angle + 1 * speed\Units  ; 1 degree per 10 ms
        ; marker offset
        x = w/2 + Sin( angle ) * radx
        y = h/2 + Sin( angle-90 ) * rady
        ; background offset
        bx = Rand( -w,w )
        by = Rand( -h,h )
        ; render marker
        SetBuffer ImageBuffer( markimg )
        ox = switch
        Repeat
            oy = switch
            Repeat
                If square
                    render = True
                Else
                    If (ox-bw/2)*(ox-bw/2)+(oy-bh/2)*(oy-bh/2) < (bh/2)*(bh/2)
                        render = True
                    Else
                        render = False
                    EndIf
                EndIf
                If render
                    c = Rand( 0,255 )
                    Color c,c,c
                    Rect ox,oy , pixw,pixh
                EndIf
                oy = oy + steph
            Until oy >= bh
            ox = ox + stepw
        Until ox >= bw
        ; draw to screen
        SetBuffer BackBuffer()
        TileImage backimg , bx,by
        DrawImage markimg , x,y
        ; info
        Color 128,192,255
        Text w/2,h/2-20 , "Frames per second: "+fps , True,True
        Text w/2,h/2-10 , "Video sync: "+Mid( "OffOn " , vsync*3+1,3 )+" <Space>" , True,True
        Text w/2,h/2    , "Time sync: "+Mid( "OffOn " , tsync*3+1,3 )+" <Enter>" , True,True
        Text w/2,h/2+10 , "Density: "+Mid( "SolidHoles" , (stepw=2)*5+1,5 )+" <Tab>" , True,True
        Text w/2,h/2+20 , "Marker: "+Mid( "Round Square" , (square)*6+1,6 )+" <Shift>" , True,True
        ; time sync
        Flip vsync
        If tsync
            WaitTimer timer
        EndIf
        switch = Not switch
    Until KeyHit( 1 )  ; Esc
    End
End Function


Function TimeSystem_DemoSimple()
    Local w = 320
    Local h = 240

    Graphics w,h , 0,3
    SetFont LoadFont( "Verdana" , 13 , True )

    Local vsync = True

    Local x# = 0
    Local incx# = 400

    Local fps
    Local frames
    Local lastsecond

    ; set 1000 ms as default speed
    Local speed.SPEED = NewSpeed( 1000 )

    SetBuffer BackBuffer()
    Repeat

        ; update time from Time System
        UpdateTime

        ; update position using speed
        x = x + incx * speed\Units  ; 400 pixels per second

        ; bounce sides
        If x <= 0
            incx = -incx
            x = 0
        ElseIf x >= w-1
            incx = -incx
            x = w-1
        EndIf

        ; draw position (vertical bar)
        Color 64,128,192
        Rect x-w/32,0 , w/32,h
        Color 255,255,255
        Rect x-w/32,0 , w/32,h , False

        ; on screen
        Color 128,192,255
        Text w/2,h/2-10 , "Frames per second: "+fps , True,True
        Text w/2,h/2+10 , "Video sync: "+Mid( "OffOn " , vsync*3+1,3 )+" <Space>" , True,True
        Flip vsync
        Cls

        ; control video sync
        If KeyHit( 57 )  ; space
            vsync = Not vsync
        EndIf

        ; measure frames per second
        frames = frames + 1
        If MilliSecs - lastsecond >= 1000
            lastsecond = MilliSecs
            fps = frames
            frames = 0
        EndIf

    Until KeyHit( 1 )  ; esc

    End
End Function


Function TimeSystem_DemoPause()
    Graphics 320,240 , 0,3
    SetFont LoadFont( "Lucida Console" , 12 , True )
    Local w = StringWidth( "W" )
    Local h = StringHeight( "|" )

    Local t$

    ; first call to match current time
    updatetime

    ; match current time
    Local basetime.TIME = newtime( 0 )

    ; 10 seconds in the past
    Local pasttime.TIME = newtime( -1000 * 10 )

    ; 10 seconds in the future
    Local fututime.TIME = newtime(  1000 * 10 )

    SetBuffer BackBuffer()
    Repeat
        updatetime

        If KeyHit( 57 )  ; space
            If basetime\paused
                resumetime basetime
            Else
                pausetime basetime
            EndIf
        EndIf

        Color 128,144,160
        Rect 0,0 , 50*w,20*h

        Color 0,0,0  :  Rect 1*w,1*h , 26*w,1*h  :  Color 0,255,0
        Text 1*w,1*h , "Current time: "+TimeSystem_DemoPause_timedisp( MilliSecs )
        If basetime\paused
            t = " <Paused>"
        Else
            t = ""
        EndIf

        Color 0,0,0  :  Rect 1*w,3*h , 35*w,2*h  :  Color 0,255,0
        Text 1*w,3*h , "Base time:    "+TimeSystem_DemoPause_timedisp( basetime\now )+t
        Text 1*w,4*h , "Press <Space>                      "

        Color 0,0,0  :  Rect 1*w,6*h , 26*w,1*h  :  Color 0,255,0
        Text 1*w,6*h , "Past time:    "+TimeSystem_DemoPause_timedisp( pasttime\now )

        Color 0,0,0  :  Rect 1*w,8*h , 26*w,1*h  :  Color 0,255,0
        Text 1*w,8*h , "Future time:  "+TimeSystem_DemoPause_timedisp( fututime\now )

        Flip
        Cls
    Until KeyHit( 1 )  ; esc

    End
End Function
; returns HH:MM:SS formatted time from specified number of milliseconds
Function TimeSystem_DemoPause_timedisp$( sex )
    Local hours
    Local mins
    Local secs
    Local msecs

    hours = sex / (60*60*1000)
    mins  = sex / (   60*1000) - hours * (60        )
    secs  = sex / (      1000) - hours * (60*60     ) - mins * (60     )
    msecs = sex                - hours * (60*60*1000) - mins * (60*1000) - secs * (1000)

    Return TimeSystem_DemoPause_string_padvalue( hours , 2 ) + ":" + TimeSystem_DemoPause_string_padvalue( mins , 2 ) + ":" + TimeSystem_DemoPause_string_padvalue( secs , 2 ) + "." + TimeSystem_DemoPause_string_padvalue( msecs , 3 )
End Function
; Add leading string to match space
; - Cuts value if too long!
Function TimeSystem_DemoPause_String_PadValue$ ( Value , Length , Fill$="0" )
    Local ValueLength

    ValueLength = Len ( Value )
    If ValueLength = Length
        Return Value  ; Keep
    Else If ValueLength > Length
        Return Right ( Value , Length )  ; Cut
    Else If ValueLength < Length
        Return String ( Fill , Length - ValueLength ) + Value  ; Add
    End If

End Function


Function TimeSystem_DemoMulti()
    Local w = 320
    Local h = 240

    Graphics w,h , 0,3
    SetFont LoadFont( "Arial" , 14 , True )
    SetBuffer BackBuffer()

    Local vsync = True 

    Local angle1# = 0
    Local angle2# = 90
    Local angle3# = 180
    Local orbit1# = 1
    Local orbit2# = h / 3
    Local orbit3# = h / 9
    Local radius1# = h / 5
    Local radius2# = radius1 / 2
    Local radius3# = radius2 / 2

    Local cx = w/2  ; screen center
    Local cy = h/2
    
    ; set orbital speeds
    Local speed1.SPEED = newspeed(  2 )
    Local speed2.SPEED = newspeed( 12 )
    Local speed3.SPEED = newspeed(  4 )

    ; create a time tracker
    Local time.TIME = newtime()
    
    ; and use this as base time for all speeds
    setspeedtime time

    Local x1,y1  ; sun
    Local x2,y2  ; planet
    Local x3,y3  ; moon

    Repeat
        updatetime

        If KeyHit( 57 )  ; space
            vsync = Not vsync
        EndIf 
        
        If KeyHit( 28 )  ; enter
            toggletime time
        EndIf 

        ; sun
        angle1 = angle1 + speed1\units
        x1 = Sin( angle1 ) * orbit1
        y1 = Sin( angle1-90 ) * orbit1

        ; planet
        angle2 = angle2 + speed2\units
        x2 = Sin( angle2 ) * orbit2
        y2 = Sin( angle2-90 ) * orbit2

        ; moon
        angle3 = angle3 - speed3\units
        x3 = Sin( angle3 ) * orbit3
        y3 = Sin( angle3-90 ) * orbit3

        ; connections
        Color 128,128,128
        Line cx+x1,cy+y1 , cx+x1+x2,cy+y1+y2
        Line cx+x1+x2,cy+y1+y2 , cx+x1+x2+x3,cy+y1+y2+y3

        ; objects
        TimeSystem_DemoMulti_DrawCircle cx+x1,cy+y1 , radius1 , 255,255,128
        TimeSystem_DemoMulti_DrawCircle cx+x1+x2,cy+y1+y2 , radius2 , 128,255,128
        TimeSystem_DemoMulti_DrawCircle cx+x1+x2+x3,cy+y1+y2+y3 , radius3 , 255,255,255

        ; info
        Color 255,0,0
        Text 0,0  , "<Enter> Pause time: "+Mid( "DisabledEnabled" , 8*time\paused+1,8 )
        Text 0,10 , "<Space> Video sync: "+Mid( "DisabledEnabled" , 8*vsync+1,8 )

        Flip
        Cls
    Until KeyHit( 1 )

    End
End Function
; Draw non-transparent outlined circle
Function TimeSystem_DemoMulti_DrawCircle( cx,cy , rad , outr=255,outg=255,outb=255 , inr=0,ing=0,inb=0 )
    Local ox,oy

    ox = cx-rad/2
    oy = cy-rad/2

    Color inr,ing,inb
    Oval ox,oy , rad,rad , True
    Color outr,outg,outb
    Oval ox,oy , rad,rad , False

End Function


;-------------------------------------------------------------------------------


; self-initialize system
TimeSystem_Init


;-------------------------------------------------------------------------------

